import platform
import os
import ctypes
import datetime
import binascii
import traceback
from ctypes import *
from functools import partial

Z_INT_RES_ID = 0
Z_STR_RES_ID = 1


Z_PUT = 0x00
Z_UPDATE = 0x01
Z_REMOVE = 0x02

Z_STORAGE_DATA = 0x00
Z_STORAGE_FINAL = 0x01
Z_EVAL_DATA = 0x02
Z_EVAL_FINAL = 0x03
Z_REPLY_FINAL = 0x04

Z_OK_TAG = 0
Z_ERROR_TAG = 1

Z_SRC_ID = 0x01
Z_SRC_SN = 0x02
Z_BRK_ID = 0x04
Z_BRK_SN = 0x08
Z_T_STAMP = 0x10
Z_KIND = 0x20
Z_ENCODING = 0x40

subscriberCallbackMap = {}
replyCallbackMap = {}
queryHandlerMap = {}
replyMap = {}


def get_lib_ext():
    system = platform.system()
    if system == 'Linux':
        return '.so'
    elif system == 'Darwin':
        return '.dylib'
    else:
        return '.dll'


def get_user_lib_path():
    system = platform.system()
    if system == 'Linux':
        return '/usr/local/lib'
    elif system == 'Darwin':
        return '/usr/local/lib'
    elif system in ['windows', 'Windows', 'win32']:
        return os.environ['ZENOH_HOME']
    else:
        return '/usr/local/lib'


system = platform.system()
if system in ['windows', 'Windows', 'win32']:
    zenoh_lib = 'zenohc' + get_lib_ext()
    zenoh_lib_path = get_user_lib_path() + os.sep + zenoh_lib
else:
    zenoh_lib = 'libzenohc' + get_lib_ext()
    zenoh_lib_path = get_user_lib_path() + os.sep + zenoh_lib


# zenoh-c result types
class z_zenoh_p_result_union_t(Union):
    _fields_ = [('zenoh', c_void_p), ('error', c_int)]


class z_zenoh_p_result_t(Structure):
    _fields_ = [('tag', c_int), ('value', z_zenoh_p_result_union_t)]


class z_pub_p_result_union_t(Union):
    _fields_ = [('pub', c_void_p), ('error', c_int)]


class z_pub_p_result_t(Structure):
    _fields_ = [('tag', c_int), ('value', z_pub_p_result_union_t)]


class z_sub_p_result_union_t(Union):
    _fields_ = [('sub', c_void_p), ('error', c_int)]


class z_sub_p_result_t(Structure):
    _fields_ = [('tag', c_int), ('value', z_sub_p_result_union_t)]


class z_sto_p_result_union_t(Union):
    _fields_ = [('sto', c_void_p), ('error', c_int)]


class z_sto_p_result_t(Structure):
    _fields_ = [('tag', c_int), ('value', z_sto_p_result_union_t)]


class z_eval_p_result_union_t(Union):
    _fields_ = [('eval', c_void_p), ('error', c_int)]


class z_eval_p_result_t(Structure):
    _fields_ = [('tag', c_int), ('value', z_eval_p_result_union_t)]


class z_vec_t(Structure):
    _fields_ = [('capacity', c_int), ('length', c_int), ('elem', c_void_p)]


# Resource Id
class z_res_id_t(Union):
    _fields_ = [('rid', c_int), ('rname', c_char_p)]


class z_resource_id_t(Structure):
    _fields_ = [('kind', c_int), ('id', z_res_id_t)]


# TimeStamp
class z_timestamp_t(Structure):
    _fields_ = [('clock_id', c_uint8 * 16),
                ('time', c_size_t)]

    def __hash__(self):
        return hash((self.time, self.clock_id[0], self.clock_id[1],
                     self.clock_id[2], self.clock_id[3], self.clock_id[4],
                     self.clock_id[5], self.clock_id[6], self.clock_id[7],
                     self.clock_id[8], self.clock_id[9], self.clock_id[10],
                     self.clock_id[11], self.clock_id[12], self.clock_id[13],
                     self.clock_id[14], self.clock_id[15]))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.time != other.time:
                return False
            for i in range(0, 15):
                if self.clock_id[i] != other.clock_id[i]:
                    return False
            return True

    def __lt__(self, other):
        if self.time < other.time:
            return True
        if self.time > other.time:
            return False
        for i in range(0, 15):
            if self.clock_id[i] < other.clock_id[i]:
                return True
            if self.clock_id[i] > other.clock_id[i]:
                return False
        return False

    def __str__(self):
        sec = self.time >> 32
        time = datetime.datetime.utcfromtimestamp(float(sec))
        frac = self.time & 0xffffffff
        ns = int((frac * 1000000000) / 0x100000000)
        id = binascii.hexlify(self.clock_id).decode("ascii")
        return "{}.{:09d}Z/{}".format(time.isoformat(), ns, id)

    def floattime(self):
        """
        Return the Timestamp's creation time as a float
        (i.e. number of seconds since Epoch:  January 1, 1970, 00:00:00 (UTC))
        Warning: the time might be rounded, depending the float precision on
        your host.
        """
        sec = self.time >> 32
        frac = self.time & 0xffffffff
        ns = float(frac) / 0x100000000
        return sec + ns

    def datetime(self, tzinfo):
        """
        Return the Timestamp's creation time as a datetime.datetime
        Warning: the time is rounded to milliseconds as datetime precision
        is millisecond.
        """
        return datetime.datetime.fromtimestamp(self.floattime(), tzinfo)


# Data Info
class z_data_info_t(Structure):
    _fields_ = [('flags', c_uint),
                ('tstamp', z_timestamp_t),
                ('encoding', c_uint8),
                ('kind', c_ushort)]


class DataInfo():
    """
    An object containing meta informations about some associated data.

    kind
        The kind of the data.
    encoding
        The encoding of the data.
    tstamp
        The unique timestamp at which the data was produced.

    """

    def __init__(self, kind=None, encoding=None, tstamp=None):
        self.kind = kind
        self.tstamp = tstamp
        self.encoding = encoding

    @staticmethod
    def from_z_data_info(z_info):
        return DataInfo(
            kind=z_info.kind if z_info.flags & Z_KIND else None,
            encoding=z_info.encoding if z_info.flags & Z_ENCODING else None,
            tstamp=z_info.tstamp if z_info.flags & Z_T_STAMP else None
        )


# Query destination
class z_query_dest_t(Structure):
    _fields_ = [('kind', c_uint8), ('nb', c_uint8)]


# Temporal properties
class z_temporal_property_t(Structure):
    _fields_ = [('origin', c_int), ('period', c_int), ('duration', c_int)]


class z_sub_mode_t(Structure):
    _fields_ = [('kind', c_uint8), ('tprop', z_temporal_property_t)]


# properties
class z_array_uint8_t(Structure):
    _fields_ = [
        ('length', c_uint),
        ('elem', POINTER(c_char))]


class z_property_t(Structure):
    _fields_ = [('id', c_size_t), ('value', z_array_uint8_t)]


def dict_to_propsvec(props):
    length = len(props)
    elems = [POINTER(z_property_t)(z_property_t(
                key, z_array_uint8_t(
                        len(val),
                        ctypes.create_string_buffer(val, len(val)))))
             for key, val in props.items()]
    return POINTER(z_vec_t)(z_vec_t(
        length, length,
        cast(((POINTER(z_property_t) * length)(*elems)), c_void_p)))


def propsvec_to_dict(vec):
    vectype = POINTER((POINTER(z_property_t) * vec.length))
    props = [prop[0] for prop in cast(vec.elem, vectype)[0]]
    return {prop.id: prop.value.elem[:prop.value.length] for prop in props}


CHAR_PTR = POINTER(c_char)


# zenoh-c callbacks
class z_reply_value_t(Structure):
    _fields_ = [
        ('kind', c_uint8),
        ('srcid', CHAR_PTR),
        ('srcid_length', c_size_t),
        ('rsn', c_uint32),
        ('rname', c_char_p),
        ('data', CHAR_PTR),
        ('data_length', c_size_t),
        ('info', z_data_info_t)]


class QueryReply(object):
    """
    An object containing one of the replies to a :func:`Zenoh.query`.

    kind
        One of the following:

        | ``Z_STORAGE_DATA`` the reply contains some data from a storage.
        | ``Z_STORAGE_FINAL`` the reply indicates that no more data is
            expected from the specified storage.
        | ``Z_EVAL_DATA`` the reply contains some data from an eval.
        | ``Z_EVAL_FINAL`` the reply indicates that no more data is expected
            from the specified eval.
        | ``Z_REPLY_FINAL`` the reply indicates that no more replies are
            expected for the query.

    source_id
        The unique identifier of the storage or eval that sent the reply
        when `kind` equals ``Z_STORAGE_DATA``, ``Z_STORAGE_FINAL``,
        ``Z_EVAL_DATA`` or ``Z_EVAL_FINAL``.
    seq_num
        The sequence number of the reply from the identified storage or
        eval when `kind` equals ``Z_STORAGE_DATA``, ``Z_STORAGE_FINAL``,
        ``Z_EVAL_DATA`` or ``Z_EVAL_FINAL``.
    rname
        The resource name of the received data when `kind` equals
        ``Z_STORAGE_DATA`` or ``Z_EVAL_DATA``.
    data
        The received data when `kind` equals ``Z_STORAGE_DATA`` or
        ``Z_EVAL_DATA``.
    info
        A :class:`DataInfo` object holding meta information about the received
        data when `kind` equals ``Z_STORAGE_DATA`` or ``Z_EVAL_DATA``.

    """

    def __init__(self, zrv):
        self.kind = zrv.kind
        self.source_id = None
        self.seq_num = None
        self.rname = None
        self.data = None
        self.info = None

        if(self.kind == Z_STORAGE_DATA
           or self.kind == Z_EVAL_DATA):
            self.source_id = zrv.srcid[:zrv.srcid_length]
            self.seq_num = zrv.rsn
            self.rname = zrv.rname.decode()
            self.data = zrv.data[:zrv.data_length]
            self.info = z_data_info_t()
            self.info.flags = zrv.info.flags
            self.info.tstamp.clock_id = zrv.info.tstamp.clock_id
            self.info.tstamp.time = zrv.info.tstamp.time
            self.info.encoding = zrv.info.encoding
            self.info.kind = zrv.info.kind

        elif(self.kind == Z_STORAGE_FINAL
             or self.kind == Z_EVAL_FINAL):
            self.source_id = zrv.srcid[:zrv.srcid_length]


class z_resource_t(Structure):
    _fields_ = [
        ('rname', c_char_p),
        ('data', c_char_p),
        ('length', c_size_t),
        ('encoding', c_ushort),
        ('kind', c_ushort)
    ]


class z_array_p_resource_t(Structure):
    _fields_ = [
        ('length', c_uint),
        ('elem', POINTER(POINTER(z_resource_t)))
    ]


ZENOH_ON_DISCONNECT_CALLBACK_PROTO = CFUNCTYPE(None, c_void_p)
ZENOH_SUBSCRIBER_CALLBACK_PROTO = CFUNCTYPE(None,
                                            POINTER(z_resource_id_t),
                                            CHAR_PTR, c_uint,
                                            POINTER(z_data_info_t),
                                            POINTER(c_int64))
ZENOH_REPLY_CALLBACK_PROTO = CFUNCTYPE(None,
                                       POINTER(z_reply_value_t),
                                       POINTER(c_int64))
ZENOH_SEND_REPLIES_PROTO = CFUNCTYPE(None,
                                     POINTER(c_int64),
                                     z_array_p_resource_t)
ZENOH_QUERY_HANDLER_PROTO = CFUNCTYPE(None,
                                      c_char_p,
                                      c_char_p,
                                      ZENOH_SEND_REPLIES_PROTO,
                                      POINTER(c_int64),
                                      POINTER(c_int64))


@ZENOH_SUBSCRIBER_CALLBACK_PROTO
def z_subscriber_trampoline_callback(rid, data, length, info, arg):
    global subscriberCallbackMap
    key = arg.contents.value
    _, callback = subscriberCallbackMap[key]
    if rid.contents.kind == Z_STR_RES_ID:
        py_info = DataInfo.from_z_data_info(info.contents)
        callback(rid.contents.id.rname.decode(), data[:length], py_info)
    else:
        print('WARNING: Received data for unknown  resource name, rid = {}'
              .format(rid.id.rid))


@ZENOH_REPLY_CALLBACK_PROTO
def z_reply_trampoline_callback(reply_value, arg):
    global replyCallbackMap
    key = arg.contents.value
    _, callback = replyCallbackMap[key]
    qr = QueryReply(reply_value.contents)
    callback(qr)


def send_replies_fun(send_replies, query_handle, replies):
    replies_array = z_array_p_resource_t()
    replies_array.length = len(replies)
    rs = (POINTER(z_resource_t) * len(replies))()
    replies_array.elem = cast(rs, POINTER(POINTER(z_resource_t)))
    i = 0
    for k, v in replies:
        d, info = v
        replies_array.elem[i].contents = z_resource_t()
        replies_array.elem[i].contents.rname = k.encode()
        replies_array.elem[i].contents.data = d
        replies_array.elem[i].contents.length = len(d)

        encoding_not_none = info is not None and info.encoding is not None
        encoding = info.encoding if encoding_not_none else 0
        replies_array.elem[i].contents.encoding = encoding

        kind_not_none = info is not None and info.kind is not None
        kind = info.kind if kind_not_none else 0
        replies_array.elem[i].contents.kind = kind

        tstamp_not_none = info is not None and info.tstamp is not None
        tstamp = info.tstamp if tstamp_not_none else 0
        replies_array.elem[i].contents.tstamp = tstamp

        i += 1
    send_replies(query_handle, replies_array)


@ZENOH_QUERY_HANDLER_PROTO
def z_query_handler_trampoline(rname,
                               predicate,
                               send_replies,
                               query_handle,
                               arg):
    global queryHandlerMap
    key = arg.contents.value
    _, handler = queryHandlerMap[key]
    try:
        handler(rname.decode(), predicate.decode(),
                partial(send_replies_fun, send_replies, query_handle))
    except Exception:
        print('WARNING: error in query handle for {} :\n{}'
              .format(rname.decode(), traceback.format_exc()))
        send_replies_fun(send_replies, query_handle, [])



from .binding import *
import socket
import time

Z_USER_KEY = 0x50
Z_PASSWD_KEY = 0x51

Z_INFO_PID_KEY = 0
Z_INFO_PEER_KEY = 1
Z_INFO_PEER_PID_KEY = 2


class SubscriberMode(object):
    """
    An object representing a subscription mode
    (see :func:`Zenoh.declare_subscriber`).

    kind
        One of the following:

        | ``SubscriberMode.Z_PUSH_MODE``
        | ``SubscriberMode.Z_PULL_MODE``
        | ``SubscriberMode.Z_PERIODIC_PUSH_MODE``
        | ``SubscriberMode.Z_PERIODIC_PULL_MODE``

    tprop
        A temporal property representing the period. `Unsupported`

    """

    Z_PUSH_MODE = 1
    Z_PULL_MODE = 2
    Z_PERIODIC_PUSH_MODE = 3
    Z_PERIODIC_PULL_MODE = 4

    def __init__(self, kind, tprop):
        self.z_sm = z_sub_mode_t()
        self.z_sm.kind = c_uint8(kind)
        self.z_sm.tprop.origin = 0
        self.z_sm.tprop.period = 0
        self.z_sm.tprop.duration = 0

        if tprop is not None:
            self.z_sm.tprop.origin = tprop.origin
            self.z_sm.tprop.period = tprop.period
            self.z_sm.tprop.duration = tprop.duration

    @staticmethod
    def push():
        """
        Return a SubscriberMode instance with `kind` =
        ``SubscriberMode.Z_PUSH_MODE``

        :returns: the equivalent of
            ``SubscriberMode(SubscriberMode.Z_PUSH_MODE, None)``
        """
        return SubscriberMode(SubscriberMode.Z_PUSH_MODE, None)

    @staticmethod
    def pull():
        """
        Return a SubscriberMode instance with `kind` =
        ``SubscriberMode.Z_PULL_MODE``

        :returns: the equivalent of
            ``SubscriberMode(SubscriberMode.Z_PULL_MODE, None)``
        """
        return SubscriberMode(SubscriberMode.Z_PULL_MODE, None)


class QueryDest(object):
    """
    An object defining which storages or evals should be destination of a
    query (see :func:`Zenoh.query`).

    kind
        One of the following:

        | ``SubscriberMode.Z_BEST_MATCH``
        | ``SubscriberMode.Z_COMPLETE``
        | ``SubscriberMode.Z_ALL``
        | ``SubscriberMode.Z_NONE``

    nb
        The number of storages or evals that should be destination of
        the query when `kind` equals ``SubscriberMode.Z_COMPLETE``.

    """
    Z_BEST_MATCH = 0
    Z_COMPLETE = 1
    Z_ALL = 2
    Z_NONE = 3

    def __init__(self, kind, nb=1):
        self.z_qd = z_query_dest_t()
        self.z_qd.kind = kind
        self.z_qd.nb = nb


def z_to_canonical_locator(locator):
    if locator is None:
        return None
    locator = locator.strip()
    a, b, c = locator.partition('/')
    if a == 'tcp' and b == '/':
        h, s, p = c.partition(':')
        if s == ':' and p != '':
            return ('tcp/' + socket.gethostbyname(h) + ':' + p).encode()
        else:
            raise Exception('Invalid locator format {}, it should be '
                            'tcp/<ip-addr|host-name>:port'.format(locator))
    elif b == '':
        h, s, p = locator.partition(':')
        if s == ':':
            return ('tcp/' + socket.gethostbyname(h) + ':' + p).encode()
        else:
            return ('tcp/' + socket.gethostbyname(h) + ':7447').encode()


class Zenoh(object):
    """
    An object that represents a zenoh session.
    """

    zenoh_native_lib = None

    def __init__(self, locator, properties={}):
        if Zenoh.zenoh_native_lib is None:
            Zenoh.zenoh_native_lib = CDLL(zenoh_lib_path)

        self.zlib = Zenoh.zenoh_native_lib

        self.zlib.z_open.restype = z_zenoh_p_result_t
        self.zlib.z_open.argtypes = [c_char_p, c_void_p, POINTER(z_vec_t)]

        self.zlib.z_info.restype = z_vec_t
        self.zlib.z_info.argtypes = [c_void_p]

        self.zlib.z_running.restype = c_int
        self.zlib.z_running.argtypes = [c_void_p]

        self.zlib.z_declare_subscriber.restype = z_sub_p_result_t
        self.zlib.z_declare_subscriber.argtypes = [
            c_void_p, c_char_p, POINTER(z_sub_mode_t),
            ZENOH_SUBSCRIBER_CALLBACK_PROTO, POINTER(c_int64)]

        self.zlib.z_declare_storage.restype = z_sto_p_result_t
        self.zlib.z_declare_storage.argtypes = [
            c_void_p, c_char_p, ZENOH_SUBSCRIBER_CALLBACK_PROTO,
            ZENOH_QUERY_HANDLER_PROTO, POINTER(c_int64)]

        self.zlib.z_declare_eval.restype = z_eval_p_result_t
        self.zlib.z_declare_eval.argtypes = [
            c_void_p, c_char_p, ZENOH_QUERY_HANDLER_PROTO,
            POINTER(c_int64)]

        self.zlib.z_declare_publisher.restype = z_pub_p_result_t
        self.zlib.z_declare_publisher.argtypes = [c_void_p, c_char_p]

        self.zlib.z_start_recv_loop.restype = c_int
        self.zlib.z_start_recv_loop.argtypes = [c_void_p]

        self.zlib.z_stop_recv_loop.restype = c_int
        self.zlib.z_stop_recv_loop.argtypes = [c_void_p]

        self.zlib.z_stream_compact_data.restype = c_int
        self.zlib.z_stream_compact_data.argtypes = [
            c_void_p, POINTER(c_char), c_int]

        self.zlib.z_stream_data_wo.restype = c_int
        self.zlib.z_stream_data_wo.argtypes = [
            c_void_p, c_char_p, c_int, c_uint8, c_uint8]

        self.zlib.z_write_data_wo.restype = c_int
        self.zlib.z_write_data_wo.argtypes = [
            c_void_p, c_char_p, c_char_p, c_int, c_uint8, c_uint8]

        self.zlib.z_pull.restype = c_int
        self.zlib.z_pull.argtypes = [c_void_p]

        self.zlib.z_query_wo.restype = c_int
        self.zlib.z_query_wo.argtypes = [
            c_void_p, c_char_p, c_char_p,
            ZENOH_REPLY_CALLBACK_PROTO, POINTER(c_int64),
            z_query_dest_t, z_query_dest_t]

        self.zlib.z_undeclare_subscriber.restype = c_int
        self.zlib.z_undeclare_subscriber.argtypes = [c_void_p]

        self.zlib.z_undeclare_storage.restype = c_int
        self.zlib.z_undeclare_storage.argtypes = [c_void_p]

        self.zlib.z_undeclare_eval.restype = c_int
        self.zlib.z_undeclare_eval.argtypes = [c_void_p]

        self.zlib.z_undeclare_publisher.restype = c_int
        self.zlib.z_undeclare_publisher.argtypes = [c_void_p]

        self.zlib.z_close.restype = c_int
        self.zlib.z_close.argtypes = [c_void_p]

        self.zlib.intersect.restype = c_int
        self.zlib.intersect.argtypes = [c_char_p, c_char_p]

        loc = z_to_canonical_locator(locator)

        r = self.zlib.z_open(loc, 0, dict_to_propsvec(properties))
        if r.tag == Z_OK_TAG:
            self.zenoh = r.value.zenoh
            self.connected = True
        else:
            raise Exception('Unable to open zenoh session (error code: {}).'
                            .format(r.value.error))

        self.zlib.z_start_recv_loop(self.zenoh)
        while not self.running:
            time.sleep(0.01)

    @staticmethod
    def open(locator=None, properties={}):
        """
        Open a zenoh session.

        :param locator: a string representing the network endpoint to which
            establish the session. A typical locator looks like this :
            ``tcp/127.0.0.1:7447``. If ``None``, :func:`open` will scout and
            try to establish the session automatically.
        :param properties: a {int: bytes} dictionary of properties that will
            be used to establish and configure the zenoh session.
            **properties** will typically contain the username and
            password informations needed to establish the zenoh session
            with a secured infrastructure. It can be set to ``NULL``.
        :returns: a handle to the zenoh session.

        """
        return Zenoh(locator, properties)

    def info(self):
        """
        Return a {int: bytes} dictionary of properties containing various
        informations about the established zenoh session.

        :returns: a {int: bytes} dictionary of properties.

        """
        return propsvec_to_dict(self.zlib.z_info(self.zenoh))

    @property
    def running(self):
        return (self.zlib.z_running(self.zenoh) != 0)

    @staticmethod
    def intersect(a, b):
        if Zenoh.zenoh_native_lib.intersect(a.encode(), b.encode()) == 1:
            return True
        else:
            return False

    def declare_publisher(self, res_name):
        """
        Declare a publication for resource **res_name**.

        :param res_name: is the resource name to publish.
        :returns: a zenoh publisher.

        """
        r = self.zlib.z_declare_publisher(self.zenoh, res_name.encode())
        if r.tag == 0:
            return r.value.pub
        else:
            raise Exception('Unable to create publisher')

    def declare_subscriber(self, selector, sub_mode, callback):
        """
        Declare a subscription for all published data matching the provided
        resource **selector**.

        :param selector: the selection of resource to subscribe to.
        :param sub_mode: the subscription mode.
        :param callback: the callback function that will be called each time a
            data matching the subscribed resource is received.
        :returns: a zenoh subscriber.

        """
        global subscriberCallbackMap
        h = hash(callback)
        k = POINTER(c_int64)(c_int64(h))
        r = self.zlib.z_declare_subscriber(self.zenoh,
                                           selector.encode(),
                                           byref(sub_mode.z_sm),
                                           z_subscriber_trampoline_callback,
                                           k)
        subscriberCallbackMap[h] = (k, callback)
        if r.tag == 0:
            return r.value.sub
        else:
            del subscriberCallbackMap[h]
            raise Exception('Unable to create subscriber')

    def declare_storage(self, selector, subscriber_callback, query_handler):
        """
        Declare a storage for all data matching the provided resource
        **selector**.

        :param selector: the selection of resources to store.
        :param subscriber_callback: the callback function that will be called
            each time a data matching the stored selection is received.
        :param query_handler: the callback function that will be called each
            time a query for data matching the stored selection is received.
            The **query_handler** function MUST call the provided
            **send_replies** function with the resulting data.
            **send_replies** can be called with an empty array.
        :returns: a zenoh storage.

        """
        global replyCallbackMap
        h = hash(query_handler)
        k = POINTER(c_int64)(c_int64(h))
        subscriberCallbackMap[h] = (k, subscriber_callback)
        queryHandlerMap[h] = (k, query_handler)
        r = self.zlib.z_declare_storage(
                self.zenoh,
                selector.encode(),
                z_subscriber_trampoline_callback,
                z_query_handler_trampoline,
                k)
        if r.tag == 0:
            return r.value.sto
        else:
            del subscriberCallbackMap[h]
            del replyCallbackMap[h]
            raise Exception('Unable to create storage')

    def declare_eval(self, selector, query_handler):
        """
        Declare an eval able to provide data matching the provided resource
        **selector**.

        :param selector: the selection of resources to evaluate.
        :param query_handler: is the callback function that will be called
            each time a query for data matching the evaluated **selector**
            is received. The **query_handler** function MUST call the provided
            **send_replies** function with the resulting data.
            **send_replies** can be called with an empty array.
        :returns: a zenoh eval.

        """
        global replyCallbackMap
        h = hash(query_handler)
        k = POINTER(c_int64)(c_int64(h))
        queryHandlerMap[h] = (k, query_handler)
        r = self.zlib.z_declare_eval(
                self.zenoh,
                selector.encode(),
                z_query_handler_trampoline,
                k)
        if r.tag == 0:
            return r.value.eval
        else:
            del replyCallbackMap[h]
            raise Exception('Unable to create eval')

    def stream_compact_data(self, pub, data):
        """
        Send data in a *compact_data* message for the resource published by
        publisher **pub**.

        :param pub: the publisher to use to send data.
        :param payload: a pointer to the data to be sent.
        :param len: the size of the data to be sent.
        :returns: 0 if the publication is successful.

        """
        self.zlib.z_stream_compact_data(pub, data, len(data))

    def stream_data(self, pub, data, encoding=0, kind=Z_PUT):
        """
        Send data in a *stream_data* message for the resource published by
        publisher **pub**.

        :param pub: the publisher to use to send data.
        :param data: the data to be sent as bytes.
        :param encoding: a metadata information associated with the published
            data that represents the encoding of the published data.
        :param kind: a metadata information associated with the published
            data that represents the kind of publication.
        :returns: 0 if the publication is successful.

        """
        self.zlib.z_stream_data_wo(pub, data, len(data), encoding, kind)

    def write_data(self, resource, data, encoding=0, kind=Z_PUT):
        """
        Send data in a *write_data* message for the resource **resource**.

        :param resource: the resource name of the data to be sent.
        :param data: the data to be sent as bytes.
        :param encoding: a metadata information associated with the published
            data that represents the encoding of the published data.
        :param kind: a metadata information associated with the published
            data that represents the kind of publication.
        :returns: 0 if the publication is successful.
        """
        self.zlib.z_write_data_wo(self.zenoh,
                                  resource.encode(),
                                  data,
                                  len(data),
                                  encoding,
                                  kind)

    def pull(self, sub):
        """
        Pull data for the `Z_PULL_MODE` or `Z_PERIODIC_PULL_MODE` subscribtion
        **sub** from the nearest infrastruture component. The pulled data will
        be provided by calling the **data_handler** function provided to the
        **declare_subscriber** function.

        :param sub: the subscribtion to pull from.
        :returns: 0 if pull is successful.
        """
        self.zlib.z_pull(sub)

    def query(self, selector, predicate, callback,
              dest_storages=QueryDest(QueryDest.Z_BEST_MATCH),
              dest_evals=QueryDest(QueryDest.Z_BEST_MATCH)):
        """
        Query data matching selection **selector**.

        :param selector: the selection of resource to query.
        :param predicate: a string that will be  propagated to the storages
            and evals that should provide the queried data. It may allow them
            to filter, transform and/or compute the queried data.
        :param reply_handler: the callback function that will be called on
            reception of the replies of the query.
        :param dest_storages: indicates which matching storages should be
            destination of the query (see :class:`QueryDest`).
        :param dest_evals: indicates which matching evals should be
            destination of the query (see :class:`QueryDest`).
        :returns: 0 if the query is sent successfully.
        """
        global replyCallbackMap
        h = hash(callback)
        k = POINTER(c_int64)(c_int64(h))
        replyCallbackMap[h] = (k, callback)
        r = self.zlib.z_query_wo(self.zenoh,
                                 selector.encode(),
                                 predicate.encode(),
                                 z_reply_trampoline_callback,
                                 k,
                                 dest_storages.z_qd,
                                 dest_evals.z_qd)
        if r != 0:
            del replyCallbackMap[h]
            raise Exception('Unable to create query')

    def undeclare_publisher(self, pub):
        """
        Undeclare the publication **pub**.

        :param pub: the publication to undeclare.
        :returns: 0 when successful.

        """
        self.zlib.z_undeclare_publisher(pub)

    def undeclare_subscriber(self, sub):
        """
        Undeclare the subscrbtion **sub**.

        :param sub: the subscription to undeclare.
        :returns: 0 when successful.

        """
        self.zlib.z_undeclare_subscriber(sub)

    def undeclare_storage(self, sto):
        """
        Undeclare the storage **sto**.

        :param sto: the storage to undeclare.
        :returns: 0 when successful.

        """
        self.zlib.z_undeclare_storage(sto)

    def undeclare_eval(self, eval):
        """
        Undeclare the eval **eval**.

        :param eval: the eval to undeclare.
        :returns: 0 when successful.

        """
        self.zlib.z_undeclare_eval(eval)

    def close(self):
        """
            Close the zenoh session.
        """
        self.zlib.z_close(self.zenoh)
        self.zlib.z_stop_recv_loop(self.zenoh)
